#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

# This shell script supports both "dw set assert" and "dw get assert"
#
# The first argument must be either "set" or "get" and this determines
# what the script does.

shopt -s extglob
TRIALS=10
LIMIT=5

declare -a active_vars

function main
{
    if [[ $# == 0 || $1 != [gs]et || ( $1 == get && $# > 2 ) || ( $1 == set && $# != 2 ) ]]
    then
	printf "\nUSAGE: dw-assert.bash {get [Adddd] | set <assertion>}\n\n"
	exit 1
    fi

    if [[ -z ${DW_INTERNAL__WORKSPACE} || -z ${DW_INTERNAL__WSPATH} ]]
    then
	printf "\nVALUE('workspace') IS UNSET\n\n"
        return 1
    fi

    get_or_set=$1
    shift

    if [[ ${get_or_set} == get ]]
    then
        local assert_label=$1

	dw_get_assert $assert_label
    elif [[ ${get_or_set} == set ]]
    then
        local assertion=$1

	if [[ -e ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/paramcurr ]]
	then
	    local discard paramcurr
	    read discard paramcurr < ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/paramcurr
	fi

	assert_label=$(dw_next_assert)

	local user_macros=${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/user-macros.m4

	[[ -e ${user_macros} ]] || user_macros=
	
	local assertion_post_m4=$(echo "${assertion}" | m4 $DWAVE_HOME/bin/dw-macros.m4 ${user_macros} -)

	assertion_post_m4=${assertion_post_m4##+([[:space:]])}
	
	dw_set_assert ${paramcurr} "${assertion_post_m4}" > ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/${assert_label} ||
	{
	    echo -e "${error_string}\n"
	    return 1
	}

	update_coeff

	{
	    display_coeff

	    invalid_qubo_operator "${assertion_post_m4}" && { echo "INVALID OPERATOR"; return; }
	    
	    is_it_linear "${assertion_post_m4}" && { echo "LINEAR"; return; } || echo "NOT LINEAR"

	    is_it_quadratic "${assertion_post_m4}" && echo "QUADRATIC" || { echo "NOT QUADRATIC"; return; }

	    (( min_obj_val == 0 )) && echo "QUBO" || echo "NOT QUBO"

	} > ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/${assert_label}.coefficient
    fi

    return $?
}

function dw_get_assert
{
    local assert_name=$1

    if [[ -n $assert_name ]]
    then
	if [[ ${assert_name} = A[0-9][0-9][0-9][0-9] && -e ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/${assert_name} ]]
	then
	    cat ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/${assert_name}
	else
	    printf "\nNO ASSERTION NAMED: ${assert_name}\n\n"
	    return 1
	fi
    else
	for assert_file in ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/A[0-9][0-9][0-9][0-9]
	do
	    if [[ ! -e ${assert_file} ]]
	    then
		printf "\nNO ACTIVE ASSERTIONS\n\n"
		return 1
	    fi

	    assert_name=${assert_file##*/}
	    awk -v assert_name=${assert_name}: '$1 ~ /^assert:$/ {sub(/^assert: '"'"'/,""); sub(/'"'"'/,""); print assert_name, $0;}' ${assert_file}
	done
    fi
}

function list_variables
{
    local assertion_expr=$1
    local pattern
    local var_list

    pattern='^([^[:alnum:]]*([[:alnum:]]*)).*'

    var_list=

    while [[ $assertion_expr =~ ${pattern} ]]
    do
	[[ -z $assertion_expr ]] && break

	assertion_expr=${assertion_expr#${BASH_REMATCH[1]}}

	[[ ${BASH_REMATCH[2]} == [[:digit:]]* ]] && continue

	var_list="$var_list ${BASH_REMATCH[2]}"
    done

    echo "${var_list}"
}

function dw_set_assert
{
    shopt -s extglob

    local constraint
    local var
    local var_list
    local token1
    local token2
    local assert_vars
    local avar
    local param_name

    if [[ $# == 1 ]]
    then
	constraint="$1"
    elif [[ $# == 2 ]]
    then
	param_name=$1
	constraint="$2"
    else
	error_string="${error_string}\nERROR: wrong number of args for dw_set_assert"
	return 1
    fi

    echo
    echo "param: ${param_name}"

    echo
    echo "assert: '${constraint}'"

    # Get the list of variables that has been declared:
    while read token1 token2
    do
	var_list="${var_list} ${token2}"
    done < ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/variable

    # Get the list of variables used in the constraint:
    assert_vars=$(list_variables "${constraint}")

    # If one of the variables used in the constraint has not been declared,
    # we have an error condition:
    error_string=
    for avar in $assert_vars
    do
	[[ "${var_list}" != !(*[[:alnum:]])${avar}!([[:alnum:]]*) ]] &&
	error_string="${error_string}\nERROR: undeclared variable \"${avar}\" appears in assert"
    done

    [[ -n $error_string ]] && return 1

    num_active_vars=0

    # Of all the declared variables, put those that are used in the
    # constraint into the active_vars[] array:
    echo
    for var in $var_list
    do
	if [[ "${constraint}" == !(*[[:alnum:]])${var}!([[:alnum:]]*) ]]
	then
	    active_vars[num_active_vars]=$var
	    num_active_vars=$(( 1 + num_active_vars ))
        fi
    done

    for ((i=0; i<num_active_vars; ++i))
    do
	echo "active_vars[$i]=${active_vars[i]}"
    done

    local header
    local format
    local dashes

    header=
    format=
    for ((i=0; i<num_active_vars; ++i))
    do
	header=" ${active_vars[i]} ${header}"
	format=" %${#active_vars[i]}d ${format}"
        dashes=" $(repeat_char ${#active_vars[i]} -) ${dashes}"
    done
    header="${header} value"
    dashes="${dashes} $(repeat_char 5 -)"
    format="${format} %5d\\n"

    echo
    echo "$header"
    echo "$dashes"

    local num_states

    num_states=$(( 2**num_active_vars ))

    local popcount
    local save_ind_1
    local save_ind_2

    for ((iloop=0; iloop<num_states; ++iloop))
    do
        state=
        i=$iloop
	popcount=0

        for ((j=0; j<num_active_vars; ++j))
        do
            lsb=$(( i % 2 ))
            i=$(( (i-lsb)/2 ))
            state="${lsb} ${state}"
            eval ${active_vars[j]}='$lsb'

	    popcount=$(( popcount + lsb ))
	    (( lsb && popcount == 1 )) && save_ind_1=$j
	    (( lsb && popcount == 2 )) && save_ind_2=$j
        done

	(( popcount == 0 )) && constant=$(( constraint ))
	(( popcount == 1 )) && linear[save_ind_1]=$(( constraint ))
	(( popcount == 2 )) && pair[num_active_vars * save_ind_2 + save_ind_1]=$(( constraint ))

        obj_val=$(( constraint ))
        valid=$(( obj_val == 0 ))
	printf "${format}" ${state} $valid

	if [[ -z $min_obj_val ]]
	then
	    min_obj_val=$obj_val
	elif (( obj_val < min_obj_val ))
	then
	    min_obj_val=$obj_val
	fi
    done

    for ((j=0; j<num_active_vars; ++j))
    do
        eval ${active_vars[j]}=0
    done

    for ((j=0; j<num_active_vars; ++j))
    do
        eval ${active_vars[j]}=2
	quadratic[j]=$(( constraint ))
        eval ${active_vars[j]}=0
    done

    echo
}

function dw_next_assert
{
    local max_assert_num
    local assert_file

    max_assert_num=0

    # We don't have to worry about resetting this to whatever the user
    # had it to since we only call this function from a subshell:

    set +o noglob

    if [[ -e ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/A0000 ]]
    then
	for assert_file in ${DW_INTERNAL__WORKSPACE}${DW_INTERNAL__WSPATH}/A[0-9][0-9][0-9][0-9]
	do
	    assert_file=10#${assert_file##*A}
	    [[ ${assert_file} > ${max_assert_num} ]] && max_assert_num=${assert_file}
	done

	max_assert_num=$(( 1 + max_assert_num ))
    fi

    printf "A%04d\n" ${max_assert_num}
}

function repeat_char
{
    local r
    local c
    local i
    local s
    r=$1
    c=$2
    for ((i=0; i<r; ++i))
    do
        s="${s}${c}"
    done
    echo $s
}

function update_coeff
{
    local i
    local j
    local a
    local b
    local d

    for ((i=0; i<num_active_vars; ++i))
    do
	a=$(( (4 * linear[i] - quadratic[i] - 3 * constant) / 2 ))
	b=$(( linear[i] - a - constant ))
	linear[i]=$a
	quadratic[i]=$b
    done

    for ((i=0; i<num_active_vars; ++i))
    do
	for ((j=i+1; j<num_active_vars; ++j))
	do
	    ind=$(( num_active_vars * j + i ))
	    d=$(( pair[ind] - linear[i] - quadratic[i] - linear[j] - quadratic[j] - constant ))
	    pair[ind]=$d
	done
    done
}

function display_coeff
{
    local i
    local j

    echo "constant: ${constant}"

    for ((j=0; j<num_active_vars; ++j))
    do
	echo "linear[$j]: ${linear[j]}"
    done

    for ((j=0; j<num_active_vars; ++j))
    do
	echo "quadratic[$j]: ${quadratic[j]}"
    done

    for ((i=0; i<num_active_vars; ++i))
    do
	for ((j=i+1; j<num_active_vars; ++j))
	do
	    ind=$(( num_active_vars * j + i ))
	    echo "pair[$i,$j]: ${pair[ind]}"
	done
    done
}

function is_it_linear
{
    local _t
    local _i
    local constraint_value
    local model
    local result=0
    local constraint=$1

    for ((_t=0; _t<TRIALS; ++_t))
    do

	for ((_i=0; _i<num_active_vars; ++_i))
	do
	    v[_i]=$(( RANDOM % LIMIT ))
            eval ${active_vars[_i]}=${v[_i]}
	done

	constraint_value=$(( constraint ))

	model=$constant
	for ((_i=0; _i<num_active_vars; ++_i))
	do
	    model=$(( model + linear[_i] * v[_i] ))
	done

	# echo "LINEAR: model=$model constraint_value=$constraint_value"

	(( model != constraint_value )) && result=1
    done

    return $result
}

function is_it_quadratic
{
    local _t
    local _i
    local _j
    local constraint_value
    local model
    local result=0
    local constraint=$1

    for ((_t=0; _t<TRIALS; ++_t))
    do

	for ((_i=0; _i<num_active_vars; ++_i))
	do
	    v[_i]=$(( RANDOM % LIMIT ))
            eval ${active_vars[_i]}=${v[_i]}
	done

	constraint_value=$(( constraint ))

	model=$constant
	for ((_i=0; _i<num_active_vars; ++_i))
	do
	    model=$(( model + linear[_i] * v[_i] ))
	    model=$(( model + quadratic[_i] * v[_i] * v[_i] ))

	    for ((_j=_i+1; _j<num_active_vars; ++_j))
	    do
		ind=$(( num_active_vars * _j + _i ))
		model=$(( model + pair[ind] * v[_i] * v[_j] ))
	    done
	done

	(( model != constraint_value )) && result=1
    done

    return $result
}

# The shell understands many operators that preclude one from
# converting an expression directly into a QUBO.  This function
# detects the presence of (some of) those operators.  The main
# function can then avoid creating a coefficient file which might
# contain misleading information.

function invalid_qubo_operator
{
    local constraint=$1

    [[ $constraint == *==* ]] && return 0
    [[ $constraint == *!=* ]] && return 0
    [[ $constraint == *++* ]] && return 0
    [[ $constraint == *--* ]] && return 0
    [[ $constraint == *\<\<* ]] && return 0
    [[ $constraint == *\>\>* ]] && return 0
    [[ $constraint == *\&* ]] && return 0
    [[ $constraint == *\|* ]] && return 0
    [[ $constraint == *^* ]] && return 0
    [[ $constraint == *\|\|* ]] && return 0
    [[ $constraint == *\&\&* ]] && return 0

    return 1
}

main "$@"
