#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

# This wrapper script reads the state file produced by "dw set assert"
# and invoked the constraint search program to see if the relation
# can be implemented by a QUBO - for relations on up to four variables.
# Since this qualifies as an "internal" routine (meaning that the
# input is the state file, which is presumed to have correct format,
# since it was created by "dw set assert") we use the exit status 0
# to convey success (i.e., the routine was able to generate a QUBO)
# or 1 to mean that no QUBO could be found.

function main
{
    a=( $(get_valid_string < $1) )

    vars=${a[0]}
    valid_string=${a[1]}

    (( vars > 4 )) && exit 1

    exec_con_search $vars $valid_string > $state_file.consearch

    found_qubo < $state_file.consearch || exit 1

    generate_posiform $state_file
}

# This function reads a state file A**** from standard input and
# writes out two pieces of information: the number of active variables
# and the bit string which represents the valid and invalid states.

function get_valid_string
{
    awk '

    BEGIN { debug = 0; varstr = ""; exit_status = 1; }

    $1 ~ /^active_vars/ {
	var_block_start = 1;
	split($0, token, "=");
	sub(/.*\[/,"",token[1]);
	sub(/\]/,"",token[1]);
	active_vars[token[1]] = token[2];
	if (0 + token[1] > 0 + max_var) max_var = token[1];
	varstr = varstr "2";
    }

    var_block_start == 1 && NF == 0 {
	var_block_start = 0;
	valid_table = NR + 3;
	dump_valid_table = 1;
	print 1+max_var;
    }

    dump_valid_table && NR >= valid_table && NF > 0 { printf("%s", $NF) }

    END { printf("\n") }

    '
}

# Set the maximum weight for the constraint search
# and then invoke the tree search which looks for QUBOs.

function exec_con_search
{
    local vars=$1
    local valid_string=$2

    [[ $vars == 2 ]] && max_weight=0
    [[ $vars == 3 ]] && max_weight=3
    [[ $vars == 4 ]] && max_weight=20

    ${DWAVE_HOME}/bin/dw-cons-search $valid_string $max_weight
}

# Parse the output of the tree search routine and look
# for the word VALID, which indicates that a QUBO was found.

function found_qubo
{
    local found_it=1
    
    while read
    do
	[[ $REPLY == *VALID* ]] && found_it=0
    done

    return $found_it
}

# Parse the output of the tree search and generate the
# output QUBO.  The output must match the A****.posiform
# file which was defined previously.

function generate_posiform
{
    local state_file=$1

    awk '

    BEGIN {
        binary[0]="2222";
        binary[1]="2221";
        binary[2]="2212";
        binary[3]="2211";
        binary[4]="2122";
        binary[5]="2121";
        binary[6]="2112";
        binary[7]="2111";
        binary[8]="1222";
        binary[9]="1221";
        binary[10]="1212";
        binary[11]="1211";
        binary[12]="1122";
        binary[13]="1121";
        binary[14]="1112";
        binary[15]="1111";
    }

    $1 ~ /^active_vars/ {
	split($0, token, "=");
	sub(/.*\[/,"",token[1]);
	sub(/\]/,"",token[1]);
	active_vars[token[1]] = token[2];
	if (0 + token[1] > 0 + max_var) max_var = token[1];
    }

    $1 ~ /^HUBO$/ && $2 ~ /^=$/ {
    hubo = $0;
    sub(/.*\[/, "", hubo);
    sub(/\]/, "", hubo);
    gsub(/ /, "", hubo);
    sub(/\r/, "", hubo);
    N = split(hubo, term, ",");
    }

    END { 
    for (i=0; i<=max_var; ++i) printf("active_var[%d]=%s\n", i, active_vars[i]);

    printf("\n");
    for (i=1; i<=N; ++i) {
        if (term[i] == 0) continue;
        w=substr(binary[i-1],4-max_var,max_var+1);
        printf("weight[\"%s\"]=%s\n", w, term[i]);
    }

    printf("\n");
    for (i=1; i<=N; ++i) {
        if (term[i] == 0) continue;
        w=substr(binary[i-1],4-max_var,max_var+1);
        printf("term: %s", term[i]);
        for (j=0; j<=max_var; ++j)
            if (substr(w,1+max_var-j,1) ~ /1/) printf("*%s", active_vars[j]);
        printf("\n");
    }

    printf("\n");
    printf("penalty: ");
    for (i=1; i<=N; ++i) {
        if (term[i] == 0) continue;
        w=substr(binary[i-1],4-max_var,max_var+1);
        printf(" + %s", term[i]);
        for (j=0; j<=max_var; ++j)
            if (substr(w,1+max_var-j,1) ~ /1/) printf("*%s", active_vars[j]);
    }
    printf("\n");

    }

    ' $state_file $state_file.consearch
}

state_file=$1

main $state_file
